
from click.exceptions import ClickException

from mvapi.cli.project import cli
from mvapi.libs.database import db
from mvapi.libs.error import save_error
from mvapi.libs.logger import init_logger
from mvapi.models import import_models
from mvapi.web.libs.appfactory import create_app
from mvapi.web.serializers import import_serializers
from mvapi.web.views import import_views


def run_app(cli_=cli):
    import_models()
    import_views()
    import_serializers()
    init_logger()

    if not cli_:
        return create_app()
    else:
        try:
            cli_.main(standalone_mode=False)
            db.session.commit()
        except ClickException as exc:
            exc.show()
        except Exception as exc:
            db.session.rollback()
            save_error()
            raise exc
        finally:
            db.session.remove()



from mvapi.libs.exceptions import AppException


class NoBlueprintException(Exception):
    pass


class NoConverterException(Exception):
    pass


class NoExtensionException(Exception):
    pass


###

class AccessDeniedError(AppException):
    pass


class AppValueError(AppException):
    pass


class BadRequestError(AppException):
    pass


class JWTError(AppException):
    pass


class NotAllowedError(AppException):
    pass


class UnauthorizedError(AppException):
    pass


class UnexpectedArgumentsError(AppException):
    pass



from mvapi.web.libs.decorators import auth_required
from mvapi.web.models.user import User
from mvapi.web.views.base import BaseView


class UsersView(BaseView):
    resource_type = 'users'
    resource_model = User

    @auth_required
    def get(self):
        meta = {}

        if not self.resource_id:
            q = self.resource_model.query.apply_args(**self.common_args)
            meta['users_count'] = q.count()

        result = self.get_resources()
        return result, meta

    def get_resource(self):
        if self.resource_id == 'me':
            if not self.current_user:
                return None

            self.resource_id = self.current_user.id_

        return super(UsersView, self).get_resource()



import click

from mvapi.web.libs.appfactory import create_app


@click.command('run', short_help='Run a development server')
@click.option('--host', '-h', help='The interface to bind to')
@click.option('--port', '-p', help='The port to bind to')
def run_(host, port):
    app = create_app()
    app.run(host=host, port=port, load_dotenv=False)



from flask import Blueprint

from mvapi.web.views.api import APIView

api_bp = Blueprint('api', __name__, url_prefix='/api')
view_func = APIView.as_view('api_view')

api_bp.add_url_rule(
    '',
    view_func=view_func
)

api_bp.add_url_rule(
    '<string:resource_type>',
    view_func=view_func
)

api_bp.add_url_rule(
    '<string:resource_type>/<string:resource_id>',
    view_func=view_func
)

api_bp.add_url_rule(
    '<string:resource_type>/<string:resource_id>/relationships/'
    '<string:relationship_type>',
    view_func=view_func
)

api_bp.add_url_rule(
    '<string:resource_type>/<string:resource_id>/'
    '<string:related_relationship_type>',
    view_func=view_func
)



import json
from collections import defaultdict, OrderedDict

from flask import g, request
from flask.views import View
from sqlalchemy import literal_column, text, union_all
from sqlalchemy.orm import Query

from mvapi.libs.database import db
from mvapi.libs.exceptions import NotFoundError
from mvapi.web.libs.misc import ApiResponse, dict_value, is_local_dev_host, \
    JSONEncoder
from mvapi.web.serializers.items import ItemsSerializer
from mvapi.web.views import RESOURCE_VIEWS


class APIView(View):
    methods = ['get', 'post', 'patch', 'delete']

    __current_user = None
    __headers = None

    def dispatch_request(self, **kwargs):
        self.__current_user = g.current_user

        req_method = request.method.lower()
        resource_type = kwargs.get('resource_type')

        if not resource_type:
            resp = ApiResponse()

        else:
            view_cls = RESOURCE_VIEWS.get(resource_type)
            if not view_cls:
                raise NotFoundError

            view = view_cls(current_user=self.__current_user, **kwargs)
            resp = view.process_request()
            if req_method == 'post':
                resp.status = 201

        is_delete = req_method == 'delete'
        results = self.__render(resp, is_delete=is_delete)

        db.session.commit()

        return results

    def __render(self, response: ApiResponse, is_delete=False):
        results = OrderedDict([
            ('links', self.__generate_response_links(response)),
        ])

        if response.meta:
            results['meta'] = response.meta

        if is_delete:
            return self.__make_response(results=results, response=response)

        relationships = self.__get_relationships(items=response.data)
        serializer = ItemsSerializer(response, relationships=relationships,
                                     current_user=self.__current_user)

        for key, value in serializer.render().items():
            results[key] = value

        return self.__make_response(results=results, response=response)

    def __generate_response_links(self, response):
        q_params = request.args.copy()

        links = OrderedDict({})
        links['self'] = self.__build_url(q_params)

        if response.limit:
            if response.current_page and response.current_page > 1:
                q_params['page[number]'] = response.current_page - 1

                if 'page[cursor]' in q_params:
                    del q_params['page[cursor]']

                links['prev'] = self.__build_url(q_params)

            if response.next_page:
                q_params['page[number]'] = str(response.next_page)

                if response.cursor:
                    q_params['page[cursor]'] = response.cursor

                links['next'] = self.__build_url(q_params)

        return links

    @staticmethod
    def __build_url(q_params):
        local_dev = is_local_dev_host()
        def_scheme = 'http' if local_dev else 'https'
        base = f'{def_scheme}://{request.host}{request.path}'

        if not q_params:
            return base

        para = '&'.join([f'{k}={v}' for k, v in q_params.items()])
        return f'{base}?{para}'

    def __make_response(self, results, response):
        self.__add_header('Content-Type', 'application/json; charset=utf-8')

        if (response.status in (201, 202,) and
                not response.relationship_type and
                not response.related_relationship_type):
            location = dict_value(results, 'data.links.self')
            if location:
                self.__add_header('Location', location)

        return (json.dumps(results, cls=JSONEncoder), response.status,
                self.__headers)

    def __add_header(self, header, value):
        if self.__headers is None:
            self.__headers = {}
        self.__headers[header] = value

    # noinspection PyMethodMayBeStatic
    def __get_relationships(self, items):
        rel_queries = defaultdict(list)
        results = {}

        if not items:
            return results

        include = request.args.get('include', '')
        if include:
            include = {i.strip() for i in include.split(',')}

        if type(items) is not list:
            items = [items]

        for item in items:
            results[str(item.id_)] = {}

            if not include:
                continue

            attr_names = set(item.available_relationships.keys())
            for attr_name in sorted(attr_names, key=lambda k: k.lower()):
                if attr_name not in include:
                    continue

                attr = getattr(item, attr_name, None)
                if not isinstance(attr, Query):
                    continue

                results[str(item.id_)][attr_name] = []

                attr = attr.add_column(
                    literal_column(f"'{item.id_}__{attr_name}'")
                    .label('relationship_for')
                )

                model = attr.column_descriptions[0]['type']
                rel_queries[model].append(attr.subquery().select())

        for model, queries in rel_queries.items():
            query = (db.session.query(model, text('relationship_for'))
                     .select_entity_from(union_all(*queries)))
            for rel_items, key in query.all():
                chunks = key.split('__')
                results[chunks[0]][chunks[1]].append(rel_items)

        return results



import os
import sys


class DefaultSettings:
    API_LOGGER_NAME = 'api'
    BLUEPRINTS = []
    CONVERTERS = []
    DEBUG = False
    DEBUG_SQL = False
    EMAILS_MODULE = None
    ENV = 'production'
    ERRORS_PATH = '.errors'
    EXTENSIONS = []
    JWTAUTH_SETTINGS = {}
    LIMIT = 15
    MIGRATIONS_EXCLUDE_TABLES = tuple()
    MODELS = []
    SERIALIZERS = []
    SQLALCHEMY_TRACK_MODIFICATIONS = False
    SYSLOG = False
    TEMPLATE_LOADERS = {}
    VIEWS = []

    # noinspection PyPep8Naming
    @property
    def LOGGING(self):
        config = {
            'version': 1,
            'disable_existing_loggers': False,
            'formatters': {
                'standard': {
                    'format': '%(asctime)s %(name)s %(levelname)s: %(message)s'
                }
            },
            'handlers': {
                'console': {
                    'formatter': 'standard',
                    'class': 'logging.StreamHandler'
                }
            },
            'loggers': {
                '': {
                    'handlers': ['console'],
                    'level': 'DEBUG' if self.DEBUG else 'INFO'
                }
            }
        }

        if self.SYSLOG:
            if type(self.SYSLOG) is not True:
                syslog_addr = self.SYSLOG
            else:
                if sys.platform == 'darwin':
                    syslog_addr = '/var/run/syslog'
                elif sys.platform == 'linux':
                    syslog_addr = '/dev/log'
                else:
                    syslog_addr = ('localhost', 514)

            config['formatters']['syslog'] = {
                'format': '%(name)s: %(message)s'
            }

            config['handlers']['syslog'] = {
                'formatter': 'syslog',
                'class': 'logging.handlers.SysLogHandler',
                'level': 'INFO',
                'address': syslog_addr
            }

        return config

    # noinspection PyPep8Naming
    @property
    def SQLALCHEMY_DATABASE_URI(self):
        return self.DB_URI

    # noinspection PyPep8Naming
    @property
    def ROOT_LOGGER_NAME(self):
        return self.APP_NAME

    def __getattr__(self, item):
        if item in os.environ:
            return os.environ[item]
        raise KeyError



from mvapi.web.libs.jsonwebtoken import JSONWebToken
from mvapi.web.models.session import Session
from mvapi.web.serializers.base import BaseSerializer


class SessionSerializer(BaseSerializer):
    resource_type = 'session'
    item: Session = None

    def render_attributes(self):
        attrs = super(SessionSerializer, self).render_attributes()
        attrs['remote_ip'] = self.item.remote_ip
        attrs['user_agent'] = self.item.user_agent

        jwt = JSONWebToken(expires_from=self.item.created_date)
        attrs['access_token'] = jwt.get_token(session=self.item)
        attrs['token_type'] = jwt.token_type
        attrs['expires'] = jwt.expires

        return attrs



import enum
import json
from datetime import date, datetime, timezone
from uuid import UUID

from flask import request, url_for as uf


class JSONEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, enum.Enum):
            return o.value
        if isinstance(o, datetime):
            return isoformat_with_timezone(o)
        if isinstance(o, date):
            return o.isoformat()
        if isinstance(o, UUID):
            return str(o)
        super(JSONEncoder, self).default(o)


def isoformat_with_timezone(timestamp):
    return timestamp.replace(tzinfo=timezone.utc).isoformat()


def is_local_dev_host():
    return '0.0.0.0' in request.host or \
           'localhost' in request.host or \
           '127.0.0.1' in request.host


def dict_value(dict_: dict, path: str, default=None):
    chunks = path.split('.')

    for idx, chunk in enumerate(chunks):
        if not isinstance(dict_, dict):
            raise KeyError

        if idx == len(chunks) - 1:
            return dict_.get(chunk, default)
        else:
            dict_ = dict_.get(chunk, {})


def url_for(endpoint, **kwargs):
    local_dev = is_local_dev_host()
    def_scheme = 'http' if local_dev else 'https'
    scheme = kwargs.get('_scheme', def_scheme)

    # Flask's url_for is pretty slow, and as long as JSON API has a certain urls
    # format we can build urls just concatenating pieces
    if endpoint == 'api.api_view':
        resource_type = kwargs.get('resource_type')
        resource_id = kwargs.get('resource_id')
        relationship_type = kwargs.get('relationship_type')
        related_relationship_type = kwargs.get('related_relationship_type')

        chunks = [request.host, 'api', resource_type, str(resource_id)]

        if relationship_type:
            chunks.append('relationships')
            chunks.append(relationship_type)

        elif related_relationship_type:
            chunks.append(related_relationship_type)

        return scheme + '://' + '/'.join(chunks)

    if '_external' not in kwargs:
        kwargs['_external'] = True

    kwargs['_scheme'] = scheme
    return uf(endpoint, **kwargs)


class ApiResponse:
    __next_page = None

    data = None
    meta = None
    status = None
    limit = 0
    current_page = 1
    cursor = None
    relationship_type = None
    related_relationship_type = None
    return_fields = None

    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            if hasattr(self, k):
                setattr(self, k, v)
            else:
                raise AttributeError

    @property
    def next_page(self):
        if type(self.data) is list:
            if self.limit and \
                    not self.__next_page and \
                    len(self.data) == self.limit:
                return self.current_page + 1

        return self.__next_page

    @next_page.setter
    def next_page(self, value):
        self.__next_page = value



import importlib
import os

from mvapi.settings import settings
from mvapi.web.views.base import BaseView

RESOURCE_VIEWS = {}


def import_views():
    views = [__package__] + settings.VIEWS

    for view_str in views:
        view = importlib.import_module(view_str)
        for file in os.listdir(os.path.dirname(view.__file__)):
            if not file.startswith('__') and file.endswith('.py'):
                name = file.rpartition('.')[0]
                mod = importlib.import_module(f'{view.__package__}.{name}')
                for attr_name in dir(mod):
                    attr = getattr(mod, attr_name)
                    if type(attr) == type and \
                            issubclass(attr, BaseView) and \
                            attr != BaseView:
                        RESOURCE_VIEWS[attr.resource_type] = attr
